module.exports = {
    sidebar: [
        {
            type: 'doc',
            id: 'introduction/introduction',
        },
        {
            type: 'doc',
            id: 'introduction/quick-start',
        },
        {
            type: 'category',
            label: 'Concepts',
            collapsed: true,
            items: [
                {
                    type: 'autogenerated',
                    dirName: '02_concepts',
                },
            ],
        },
        {
            type: 'category',
            label: 'Guides',
            collapsed: true,
            items: [
                {
                    type: 'autogenerated',
                    dirName: '03_guides',
                },
            ],
        },
        {
            type: 'category',
            label: 'Upgrading',
            collapsed: true,
            items: [
                {
                    type: 'autogenerated',
                    dirName: '04_upgrading',
                },
            ],
        },
        {
            type: 'doc',
            id: 'changelog',
        },
    ],
};
