import OriginalComponentTypes from '@theme-original/NavbarItem/ComponentTypes';
import VersionedReferenceNavbarItem from './VersionedReferenceNavbarItem';

export default {
    ...OriginalComponentTypes,
    'custom-versioned-reference': VersionedReferenceNavbarItem,
};
