/**
 * @type {import('prettier').Options}
 */
module.exports = {
    parser: 'markdown',
    arrowParens: 'avoid',
    trailingComma: 'all',
    singleQuote: true,
    tabWidth: 4,
    printWidth: 150,
    proseWrap: 'always',
};
