import React from 'react';

export default function Gradients() {
    return (
        <svg xmlns="http://www.w3.org/2000/svg" width="0" height="0" viewBox="0 0 0 0" fill="none">
            <defs>
                <linearGradient id="gradient-1" x1="26.6667" y1="12" x2="14.2802" y2="34.5208"
                                gradientUnits="userSpaceOnUse">
                    <stop offset="0%" stop-color="#9dceff"/>
                    <stop offset="70%" stop-color="#4584b6"/>
                    <stop offset="100%" stop-color="#4584b6"/>
                </linearGradient>
                <linearGradient id="gradient-2" x1="29.6667" y1="0" x2="-1.80874" y2="26.2295"
                                gradientUnits="userSpaceOnUse">
                <stop offset="0%" stop-color="#4584b6"/>
                </linearGradient>
            </defs>
        </svg>
    );
}
